from datetime import datetime
from typing import Optional

from flaschengeist.models.user import User

from flaschengeist.database import db
from flaschengeist.models import ModelSerializeMixin, UtcDateTime


class Transaction(db.Model, ModelSerializeMixin):
    __tablename__ = "balance_transaction"
    # Protected foreign key properties
    _receiver_id = db.Column("receiver_id", db.Integer, db.ForeignKey("user.id"))
    _sender_id = db.Column("sender_id", db.Integer, db.ForeignKey("user.id"))
    _author_id = db.Column("author_id", db.Integer, db.ForeignKey("user.id"), nullable=False)
    _reversal_id = db.Column("reversal_id", db.Integer, db.ForeignKey("balance_transaction.id"))

    # Public and exported member
    id: int = db.Column("id", db.Integer, primary_key=True)
    time: datetime = db.Column(UtcDateTime, nullable=False, default=UtcDateTime.current_utc)
    amount: float = db.Column(db.Numeric(precision=5, scale=2, asdecimal=False), nullable=False)
    reversal: Optional["Transaction"] = db.relationship("Transaction", foreign_keys=[_reversal_id])

    # Dummy properties used for JSON serialization (userid instead of full user)
    sender_id: Optional[str] = ""
    receiver_id: Optional[str] = ""
    author_id: Optional[str] = ""

    # Not exported relationships just in backend only
    sender_: User = db.relationship("User", foreign_keys=[_sender_id])
    receiver_: User = db.relationship("User", foreign_keys=[_receiver_id])
    author_: User = db.relationship("User", foreign_keys=[_author_id])

    @property
    def sender_id(self):
        return self.sender_.userid if self.sender_ else None

    @property
    def receiver_id(self):
        return self.receiver_.userid if self.receiver_ else None

    @property
    def author_id(self):
        return self.author_.userid

    # Override to prevent circular dependencies (endless JSON)
    def serialize(self):
        d = super().serialize()
        if d["reversal"]:
            d["reversal"].reversal = None
        else:
            d["reversal"] = None
        return d
